from __future__ import annotations
import numpy as np
import pandas as pd
from pathlib import Path
from typing import List, Optional
from saltysplits.enums import TimeType
from saltysplits.annotations import encode_time
from saltysplits.models import Splits


class SaltySplits(Splits):
    """
    Main interface for deserializing LiveSplit files (LSS) and interacting with the speedrunning data within.

    Args:
        Splits (Splits): Root pydantic-xml model, parent to all other LSS elements and attributes

    Returns:
        SaltySplits: Instanced pydantic-xml model containing deserialized speedrunning data and formatting functionality
    """

    @classmethod
    def read_lss(cls, lss_path: Path) -> SaltySplits:
        """
        Reads a LiveSplit file (LSS) as a SaltySplits root model instance (and populate it with validated splits, runs, attempts and segments).

        Args:
            lss_path (Path): Path to the LiveSplit file (.LSS)

        Returns:
            SaltySplits: Instanced pydantic-xml model containing deserialized speedrunning data and formatting functionality
        """

        with open(lss_path, "rb") as file:
            xml_bytes = file.read()
        return cls.from_xml(xml_bytes)

    def _collect_ids(self) -> List[Optional[str]]:
        """
        Iterates over all splits and attempts to find all unique run IDs (including those from deleted or otherwise partial runs)

        Returns:
            List[Optional[str]]: A list of unique run IDs (sorted as integers if possible)
        """

        run_ids = set()
        run_ids.update([attempt.id for attempt in self.attempt_history])
        run_ids.update([split.id for segment in self.segments for split in segment.segment_history])  # fmt: skip
        run_ids = list(run_ids)

        # sort them as int if possible
        if all(map(lambda x: str(int(x)) == x, run_ids)):
            run_ids = sorted(run_ids, key=lambda x: int(x))
        return run_ids

    def is_comparable(self, other: SaltySplits, strict: bool = True) -> bool:
        """
        Determines whether two SaltySplits instances pertain to the same topic.
        This is done by comparing their GameName / CategoryName values and the number of
        segments (and their exact Segment.Name values if strict is set)

        Args:
            other (SaltySplits): Another SaltySplits instance to compare against
            strict (bool, optional): Whether segment names need to match. Defaults to True.

        Returns:
            bool: True if runs from given instances can be compared, False if not
        """

        same_game = self.game_name == other.game_name
        same_category = self.category_name == other.category_name
        same_count = len(self.segments) == len(other.segments)
        same_segments = all([self.segments[i].name == other.segments[i].name for i in range(len(self.segments))]) if same_count and strict else not strict  # fmt: skip
        return all([same_game, same_category, same_count, same_segments])

    def to_df(
        self,
        time_type: TimeType = TimeType.REAL_TIME,
        allow_partial: bool = False,
        allow_empty: bool = False,
        cumulative: bool = False,
        lss_repr: bool = False,
        lss_ns: bool = True,
    ) -> pd.DataFrame:
        """
        Iterates over all splits to reconstruct individual runs (partial or otherwise) and represent them as a single pandas.DataFrame.
        Can then be used for further analysis within Python or dumped to common formats for use outside (e.g. see pandas.DataFrame.to_csv)

        Args:
            time_type (TimeType, optional): Whether to use GameTime or RealTime values. Defaults to REAL_TIME.
            allow_partial (bool, optional): Whether to allow runs that don't have values for all segments. Defaults to False.
            allow_empty (bool, optional): Whether to allow runs that don't have values for any segments. Defaults to False.
            cumulative (bool, optional): Whether succesive splits in a run have to add up to the total runtime. Defaults to False.
            lss_repr (bool, optional): Whether to use LSS' string representation of time (e.g. "01:55:11.1422649") or to keep the pandas.TimeDelta representation. Defaults to False.
            lss_ns (bool, optional): Whether you want to include nanoseconds in LSS' string representation of time. Defaults to True.

        Returns:
            pd.DataFrame: pandas.DataFrame of shape (n_segments, n_runs) containing run data
        """

        run_ids = self._collect_ids()
        segment_names = [segment.name for segment in self.segments]
        placeholder_data = np.full((len(segment_names), len(run_ids)), np.timedelta64("NaT", "ns"))
        placeholder_df = pd.DataFrame(placeholder_data, columns=run_ids, index=segment_names)

        for segment in self.segments:
            for split in segment.segment_history:
                if time_type == time_type.GAME_TIME:
                    time = split.game_time
                elif time_type == time_type.REAL_TIME:
                    time = split.real_time
                placeholder_df.loc[segment.name, split.id] = time

        if not allow_partial:
            # drops runs that have one or more NaT values (i.e. incomplete runs)
            placeholder_df = placeholder_df.loc[:, ~placeholder_df.isna().any()]

        if not allow_empty:
            # drops runs that only have NaT values (i.e. empty runs)
            placeholder_df = placeholder_df.loc[:, ~placeholder_df.isna().all()]

        if cumulative:
            # skipna ensures that we stop accumulating if we miss a split in beween
            placeholder_df = placeholder_df.apply(lambda column: column.cumsum(skipna=False))

        if lss_repr:
            # formats timedelta values to same string representation as used in LSS files
            placeholder_df = placeholder_df.map(
                lambda x: encode_time(x, include_ns=lss_ns) if pd.notna(x) else None
            )

        return placeholder_df



import pytest
import requests
import pathlib
from lxml.etree import Element

# names come from livesplit-core/tests/run_files/mod.rs
LSS_DIR = pathlib.Path(__file__).parent / "run_files"


def drop_empty_tags(element: Element, top_level=True) -> None:
    """Removes empty XML tags from a given LXML ElementTree

    Args:
        element (Element): Root XML element from which empty elements should be removed
        top_level (bool, optional): Whether to only check direct children or all children (recursively)
    """
    # dropping empty tags before comparison, no way to catch that AND actually optional elements (e.g. real_time and/or game_time)
    children = list(element) if top_level else list(element.iter())
    for child in children:
        if len(child) == 0 and not child.text and not child.attrib:
            parent = child.getparent()
            if parent is not None:
                parent.remove(child)


def get_run_files(lss_dir: pathlib.Path) -> None:
    api_url = "https://api.github.com/repos/LiveSplit/livesplit-core/contents/tests/run_files"
    response = requests.get(api_url)
    files = response.json()

    lss_files = [file for file in files if file["name"].endswith(".lss")]
    for lss_file in lss_files:
        lss_src = lss_file["download_url"]
        lss_dst = lss_dir / lss_file["name"]

        if lss_dst.exists() and lss_dst.stat().st_size == lss_file["size"]:
            continue

        with requests.get(lss_src, stream=True) as response:
            response.raise_for_status()
            with open(lss_dst, "w", encoding="utf-8") as file:
                file.write(response.text)


@pytest.fixture
def LIVESPLIT_1_7_0():
    return LSS_DIR / "Grand Theft Auto Vice City - Any% (No SSU).lss"


@pytest.fixture
def LIVESPLIT_1_3():
    return LSS_DIR / "livesplit1.5.lss"



import re
from pandas import Timedelta
from datetime import datetime
from pydantic import BeforeValidator, PlainSerializer, AfterValidator
from packaging.version import Version
from typing import Optional, Annotated, List
from saltysplits.constants import (
    DATETIME_FORMAT,
    NANOSECONDS_DAY,
    NANOSECONDS_HOUR,
    NANOSECONDS_MINUTE,
    NANOSECONDS_SECOND,
    MINIMUM_LSS_VERSION,
)


def decode_time(value: str) -> Timedelta:
    """
    Decodes a string using LSS` time representation as a pandas.Timedelta object (including optional days prefix)

    Args:
        value (str): String using LSS' time representation (e.g. "1.01:55:11.1422649")

    Returns:
        Timedelta: Timedelta representation of the input string
    """

    pattern = re.compile(
        r"^(?:(?P<days>\d+)\.)?(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)(?:\.(?P<fraction>\d+))?$"
    )
    match = pattern.match(value)
    assert match, (
        "Invalid time format, expected 'HH:MM:SS' (with optional days prefix and fraction suffix)"
    )
    groups = match.group("days", "hours", "minutes", "seconds", "fraction")
    days, hours, minutes, seconds, fraction = map(lambda x: int(x) if x else 0, groups)
    nanoseconds = fraction * 100

    return Timedelta(
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        nanoseconds=nanoseconds,
    )


def parse_timedelta(timedelta: Timedelta) -> List[int]:
    """
    Decomposes a Timedelta object into its equivalent days, hours, minutes, seconds
    and nanoseconds. Divmod is used to prevent floating-point errors

    Args:
        timedelta (Timedelta): Timedelta object to decompose

    Returns:
        List[int]: Integer list containing days, hours, minutes, seconds and nanoseconds
    """

    days, remainder = divmod(timedelta.value, NANOSECONDS_DAY)
    hours, remainder = divmod(remainder, NANOSECONDS_HOUR)
    minutes, remainder = divmod(remainder, NANOSECONDS_MINUTE)
    seconds, remainder = divmod(remainder, NANOSECONDS_SECOND)
    nanoseconds = remainder // 100
    return days, hours, minutes, seconds, nanoseconds


def encode_time(content: Timedelta, include_ns: bool = True) -> str:
    """
    Encodes a Timedelta object as the same string format used in the LSS files (i.e. 'HH:MM:SS' with optional days prefix and fractional seconds suffix)

    Args:
        content (Timedelta): Timedelta object to represent as string using LSS' time representation
        include_ns (bool, optional): Whether to include nanoseconds in the formatted string. Defaults to True.

    Returns:
        str: LSS string representation of input Timedelta object
    """

    # optionally adds n_nanoseconds suffix but only adds n_days prefix if not 0
    days, hours, minutes, seconds, nanoseconds = parse_timedelta(timedelta=content)
    delta_string = f"{hours:02}:{minutes:02}:{seconds:02}"
    delta_string = f"{delta_string}.{nanoseconds:07}" if include_ns else delta_string
    delta_string = f"{days}.{delta_string}" if days else delta_string
    return delta_string


def encode_offset(content: Timedelta) -> str:
    """
    Encodes a Timedelta object as the same string format used in the LSS files (i.e. 'HH:MM:SS' with optional days prefix and fractional seconds suffix)

    Args:
        content (Timedelta): Timedelta object to represent as string using LSS' time representation
        include_ns (bool, optional): Whether to include nanoseconds in the formatted string. Defaults to True.

    Returns:
        str: LSS string representation of input Timedelta object
    """

    # only adds n_nanoseconds suffix and n_days prefix if not 0
    days, hours, minutes, seconds, nanoseconds = parse_timedelta(timedelta=content)
    delta_string = f"{hours:02}:{minutes:02}:{seconds:02}"
    delta_string = f"{delta_string}.{nanoseconds:07}" if nanoseconds else delta_string
    delta_string = f"{days}.{delta_string}" if days else delta_string
    return delta_string


def decode_datetime(value: str) -> datetime:
    return datetime.strptime(value, DATETIME_FORMAT)


def encode_datetime(content: datetime) -> str:
    return content.strftime(DATETIME_FORMAT)


def verify_version(lss_version: Version) -> Version:
    if lss_version < MINIMUM_LSS_VERSION:
        raise ValueError(
            f"Minimum LSS file version is {MINIMUM_LSS_VERSION}. Consider exporting your {lss_version} LSS file with LiveSplit to reencode it to a more recent version"
        )
    return lss_version


TimeOptional = Annotated[
    Optional[Timedelta],
    PlainSerializer(encode_time, when_used="unless-none"),
    BeforeValidator(decode_time),
]

OffsetOptional = Annotated[
    Optional[Timedelta],
    PlainSerializer(encode_offset, when_used="unless-none"),
    BeforeValidator(decode_time),
]

DateTime = Annotated[
    datetime,
    PlainSerializer(encode_datetime, when_used="unless-none"),
    BeforeValidator(decode_datetime),
]


SBool = Annotated[
    bool,
    PlainSerializer(lambda x: str(x), when_used="unless-none"),
]

LssVersion = Annotated[
    Version,
    BeforeValidator(Version),
    AfterValidator(verify_version),
    PlainSerializer(lambda x: str(x), when_used="always"),
]



import numpy as np
import re
import pandas as pd
from pathlib import Path
import streamlit as st
import saltysplits as ss
import altair as alt
from typing import List
from pydantic import ValidationError
from saltysplits.annotations import encode_time
from saltysplits.constants import MINIMUM_LSS_VERSION
from saltysplits import SaltySplits
from saltysplits import DEMO_SPLITS
from saltysplits import TimeType


st.set_page_config(
    page_title="Sup! - SaltySplits.com",
    page_icon="🧂",
    initial_sidebar_state="expanded",
    layout="wide",
    menu_items={
        'Get help': 'https://github.com/jaspersiebring/saltysplits',
        'Report a bug': "https://github.com/jaspersiebring/saltysplits/issues/new",
    }
)

@st.cache_data
def represent_resets(splits_df: pd.DataFrame) -> pd.DataFrame:
    reset_indices = np.argmax(splits_df.isna(), axis=0)
    reset_segments = np.array(splits_df.index)[reset_indices]
    reset_segment_names, reset_segment_counts = np.unique(reset_segments, return_counts=True)
    reset_df = pd.DataFrame({"Segment": reset_segment_names, "Count": reset_segment_counts})
    return reset_df

@st.cache_data
def represent_time(td_series: pd.Series, include_ns: bool = False) -> pd.DataFrame:
    dt_series = pd.to_datetime(td_series, unit="ns")  
    ht_series = td_series.apply(lambda x: encode_time(x, include_ns=include_ns) if pd.notna(x) else None) 
    time_dataframe = pd.concat([ht_series, dt_series], axis=1) 
    time_dataframe.columns = ["Time", "DateTime"]
    time_dataframe = time_dataframe.rename_axis("id").reset_index()
    return time_dataframe

@st.cache_data
def splits_from_bytes(lss_bytes: bytes) -> SaltySplits:
    return SaltySplits.from_xml(lss_bytes)

@st.cache_data
def splits_dataframe(splits_bytes: bytes, time_type: TimeType, allow_partial: bool, allow_empty: bool = False, cumulative: bool = False, lss_repr: bool = False, lss_ns: bool = True) -> pd.DataFrame:
    # mainly exists because we can't cache SaltySplits instances
    splits = splits_from_bytes(splits_bytes)
    dataframe = splits.to_df(
        time_type=time_type,
        allow_partial=allow_partial,
        allow_empty=allow_empty,
        cumulative=cumulative,
        lss_repr=lss_repr, 
        lss_ns=lss_ns
    )
    return dataframe
    
@st.cache_data
def read_bytes(path: Path) -> bytes:
    with open(path, "rb") as file:
        return file.read()


@st.cache_data
def splits_metrics(splits_bytes: bytes, time_type: TimeType) -> None:
    splits = splits_from_bytes(splits_bytes)
    splits_df = splits_dataframe(splits_bytes, time_type=time_type, allow_partial=True)

    complete_runs = splits_df.loc[:, ~splits_df.isna().any()]  
    run_times = complete_runs.sum(axis=0)
    best_run_index = run_times.argmin()
    best_run = run_times.iloc[best_run_index]
    best_run = encode_time(best_run, include_ns=False)

    best_segments = splits_df.min(axis=1)
    best_segments_sum = best_segments.sum()
    best_segments_sum = encode_time(best_segments_sum, include_ns=False)
    life_playtime = splits_df.sum(axis=0).sum()
    life_playtime = encode_time(life_playtime, include_ns=False)
    best_diffs = complete_runs.iloc[:, best_run_index] - best_segments
    possible_timesave = best_diffs.sum()
    possible_timesave = encode_time(possible_timesave, include_ns=False)

    time_column, segment_column, timesave_column, attempts_column, playtime_column = st.columns(5)
    time_column.metric("BEST TIME", best_run)
    segment_column.metric("SUM OF BEST SEGMENTS", best_segments_sum)
    timesave_column.metric("POSSIBLE TIMESAVE", possible_timesave)
    attempts_column.metric("ATTEMPTS", f"{splits.attempt_count}")
    playtime_column.metric("LIFE PLAYTIME", life_playtime)

@st.cache_data
def rank_splits(dataframe, n: int = 3) -> List[str]:
    # LEGACY (add_suffix)
    ranked_indices = []
    for name, group in dataframe.T.groupby(lambda x: re.findall(r'\((.*)\)', x)[0]):
        complete_group = group.T.loc[:, ~group.T.isna().any()] 
        ranked_indices.extend(complete_group.sum(axis=0).sort_values().head(n).index.tolist())
    return ranked_indices


def resets_piechart(
    dataframe: pd.DataFrame,
    title: str = "Resets Per Segment",
) -> None:

    highlight = alt.selection_point(
        fields=["Segment"],
        on="mouseover",
        clear="mouseout",
        empty="all",
        bind="legend"
    )

    chart = alt.Chart(
        dataframe,
        title=alt.Title(title, anchor="middle")
    ).mark_arc(
        innerRadius=50,
        cursor="pointer",
        stroke="white",
        strokeWidth=2
    ).encode(
        theta=alt.Theta(field="Count", type="quantitative"),
        color=alt.Color(
            field="Segment",
            type="nominal",
            legend=alt.Legend(
                symbolSize=200,
                symbolStrokeColor=None,
                symbolStrokeWidth=0
            )
        ),
        opacity=alt.condition(highlight, alt.value(1), alt.value(0.3))
    ).add_params(highlight)

    st.altair_chart(chart)


def time_linegraph(
    dataframe: pd.DataFrame,
    title: str,
    x_title: str,
    y_title: str,
    interpolation_method: str = "monotone",
) -> None:

    nearest = alt.selection_point(
        nearest=True, on="pointerover", fields=["id"], empty=False, clear="pointerout"
    )

    line_graph = alt.Chart(
        dataframe,
        title=alt.Title(title, anchor="middle")
    ).mark_line(
        interpolate=interpolation_method,
        orient="horizontal",
        point=True
    ).encode(
        x=alt.X("id:Q", axis=alt.Axis(grid=False, title=x_title)),
        y=alt.Y("DateTime:T", axis=alt.Axis(format="%H:%M:%S", grid=True, title=y_title)),
        order=alt.Order(field="id:Q"),
        tooltip=["id", "Time"],
    )

    points = line_graph.mark_point().encode(
        opacity=alt.condition(nearest, alt.value(1), alt.value(0))
    )

    rules = alt.Chart(dataframe).mark_rule(color="gray").encode(
        x="id:Q",
        opacity=alt.condition(nearest, alt.value(0.3), alt.value(0)),
        tooltip=["id", "Time"]
    ).add_params(nearest)

    layer = alt.layer(line_graph, points, rules)
    st.altair_chart(layer)


if __name__ == "__main__":
    with st.sidebar:
        LOGO_URL = "https://github.com/user-attachments/assets/5863a3ed-0552-449f-8942-4378fbd1a59e"
        st.markdown(
            f"""
            <div style='display: flex; align-items: center;'>
                <img src='{LOGO_URL}' style='width: 200px; height: 200px;'>
            </div>
            """,
            unsafe_allow_html=True,
        )
        
        lss_file = st.file_uploader(
            f"Upload LSS file (minimum supported version is {MINIMUM_LSS_VERSION})",
            type=["lss", "xml"],
            accept_multiple_files=False,
            label_visibility="visible"
        )
        
        if lss_file is not None:
            lss_bytes = lss_file.getvalue()
            try:
                if splits_dataframe(lss_bytes, time_type=TimeType.REAL_TIME, allow_partial=False).empty:
                    st.error(f"No runs with {TimeType.REAL_TIME.name} values found, defaulting to example LSS file")
                    lss_bytes = read_bytes(DEMO_SPLITS)    
                else:
                    st.toast("Did you know that you can export your Run Stats? Hover the table and click 'Download as CSV'")
                    
            except ValidationError as e:
                st.warning(f"{e}")
                st.toast(f"{e}")
        else:
            lss_bytes = read_bytes(DEMO_SPLITS)
        splits = splits_from_bytes(lss_bytes)

    # LIFETIME STATS
    st.title(f"{splits.game_name} / :primary[{splits.category_name}]")
    splits_metrics(splits_bytes=lss_bytes, time_type=TimeType.REAL_TIME)

    for i, tab in enumerate(st.tabs(["Real Time", "Game Time"])):
        st.session_state.time_type = TimeType(i)
        complete_runs = splits_dataframe(lss_bytes, time_type=st.session_state.time_type, allow_partial=False)
        partial_runs = splits_dataframe(lss_bytes, time_type=st.session_state.time_type, allow_partial=True)
        cumulative_runs = splits_dataframe(lss_bytes, time_type=st.session_state.time_type, allow_partial=True, cumulative=True)

        with tab:
            with st.container(border=True):
                if not complete_runs.empty:
                    lifetime_line, resets_pie = st.columns(2, gap="large")
                    with lifetime_line:
                        run_times = complete_runs.sum(axis=0)
                        run_times_df = represent_time(run_times, include_ns=False)
                        time_linegraph(run_times_df, title="Completed Run Duration over Time", x_title="Attempt Number", y_title = "Run Duration")
                    
                    with resets_pie:
                        resets_df = represent_resets(partial_runs)
                        resets_piechart(resets_df)
                else:
                    st.error(f"No runs with {st.session_state.time_type.name} values found, Lifetime Stats visualization not possible")

                            
            st.subheader("Run Stats", divider="blue")
            with st.container(border=True):
                if not partial_runs.empty:
                    n = min(3, len(complete_runs.columns))
                    segment_order = complete_runs.index.to_list()
                    default_ids = complete_runs.sum(axis=0).sort_values().head(n).index.to_list()
                    run_ids = st.multiselect(f"Select run (includes top {n} runs by default)", options= partial_runs.columns.to_list(), key=f"multirun_selector_{st.session_state.time_type.name}", default=default_ids)
                    
                    run_table, run_graph = st.columns(2, gap="large")
                    with run_table.container():
                        if f"pills_{st.session_state.time_type.name}" not in st.session_state:
                            st.session_state[f"pills_{st.session_state.time_type.name}"] = []
                            
                        format_options = st.session_state[f"pills_{st.session_state.time_type.name}"]
                        cumulative = "Cumulative" in format_options
                        nanoseconds = "Nanoseconds" in format_options
                        formatted_runs = splits_dataframe(lss_bytes, time_type=st.session_state.time_type, allow_partial=True, cumulative=cumulative, lss_ns=nanoseconds, lss_repr=True)
                        
                        selected_runs = formatted_runs.loc[:, run_ids]
                        st.dataframe(selected_runs, height=len(segment_order) * 30, key=f"table_{st.session_state.time_type.name}")
                        format_options = st.pills("Format options", options = ["Cumulative", "Nanoseconds"], key=f"pills_{st.session_state.time_type.name}", selection_mode = "multi", default=[])

                    with run_graph:                       
                        selected_lines = cumulative_runs.loc[:, run_ids]
                        selected_lines = pd.melt(selected_lines.T.rename_axis("id").reset_index(), id_vars='id', value_name="TimeDelta", var_name="Segment")
                        selected_lines["Time"] = selected_lines["TimeDelta"].apply(lambda x: encode_time(x, include_ns=False) if pd.notna(x) else None) 
                        selected_lines["DateTime"] = pd.to_datetime(selected_lines["TimeDelta"], unit="ns")

                        line_graph = alt.Chart(selected_lines, title=alt.Title("Run Breakdown (Cumulative)", anchor="middle")).mark_line(
                            point=True,
                            interpolate="monotone",
                            orient="horizontal").encode(
                                alt.Y("Segment:N", axis=alt.Axis(grid=True)).title("Segment Name").sort(segment_order),
                                alt.X("DateTime:T", axis=alt.Axis(grid=False, format="%H:%M:%S")).title("Run Duration"),
                                alt.Order(field="DateTime"),
                                color='id:N',
                                tooltip=["Time", "Segment", "id"]
                            ).properties(
                                height=len(segment_order) * 30
                            )
                        st.altair_chart(line_graph)
  
                    st.divider()
                    
                    selected_bars = partial_runs.loc[:, run_ids]
                    selected_bars = pd.melt(selected_bars.T.rename_axis("id").reset_index(), id_vars='id', value_name="TimeDelta", var_name="Segment")
                    selected_bars["Time"] = selected_bars["TimeDelta"].apply(lambda x: encode_time(x, include_ns=nanoseconds) if pd.notna(x) else None) 
                    selected_bars["DateTime"] = pd.to_datetime(selected_bars["TimeDelta"], unit="ns")
                    selected_bars["Zero"] = pd.to_datetime(0, unit="ns")

                    bar_graph = alt.Chart(selected_bars, title = alt.Title("Run Breakdown (Segmented)", anchor="middle")).mark_bar().encode(
                        alt.Y("DateTime:T", axis=alt.Axis(grid=True, format="%H:%M:%S")).title("Segment Duration"),
                        alt.X("Segment:N", axis=alt.Axis(labelAngle=-90)).title("Segment Name").sort(segment_order),
                        y2="Zero:T",
                        xOffset="id:N",
                        color="id:N"
                    )
                    st.altair_chart(bar_graph)

                else:
                    st.error(f"No splits with {st.session_state.time_type.name} values found, Run Stats visualization not possible")
                    

            st.subheader("Segment Stats", divider="blue")
            
            segment_name = st.selectbox("Select segment (defaults to first)", list(partial_runs.index), 0, key=f"segment_selector_{st.session_state.time_type.name}")
            segment_times = partial_runs.loc[segment_name, :]
            segment_times = segment_times.loc[~segment_times.isna()] 
        
            with st.container(border=False):
                if not segment_times.empty:
                    with st.container(border=True):
                        segment_stats = represent_time(segment_times, include_ns=False)
                        time_linegraph(dataframe=segment_stats, title=f"Segment Duration over Time ({segment_name})", x_title = "Attempt Number", y_title="Segment Duration")

                    with st.container(border=True):
                        min_segment, max_segment, mean_segment, median_segment, std_segment = st.columns(5)
                        min_segment.metric(f"MIN (run {segment_times.index[segment_times.argmin()]})", encode_time(segment_times.min(), include_ns=False))
                        max_segment.metric(f"MAX (run {segment_times.index[segment_times.argmax()]})", encode_time(segment_times.max(), include_ns=False))
                        mean_segment.metric("MEAN", encode_time(segment_times.mean(), include_ns=False))
                        median_segment.metric("MEDIAN", encode_time(segment_times.median(), include_ns=False))
                        std_segment.metric("STANDARD DEVIATION", f"{int(segment_times.std().total_seconds())} seconds")
                else:
                    st.error(f"No splits with {st.session_state.time_type.name} values for segment {segment_name} found, Segment Stats visualization not possible")



import pytest
import numpy as np
import lxml.etree as ET
from pydantic import ValidationError
from saltysplits import SaltySplits as ss
from saltysplits.enums import TimeType


class TestSaltySplits:
    def test_read_lss(self, LIVESPLIT_1_7_0):
        tree = ET.parse(LIVESPLIT_1_7_0)
        element = tree.getroot()
        model_from_tree = ss.from_xml_tree(element)
        model_from_lss = ss.read_lss(LIVESPLIT_1_7_0)
        assert model_from_tree == model_from_lss

    def test_read_older_lss(self, LIVESPLIT_1_3):
        with pytest.raises(ValidationError):
            ss.read_lss(LIVESPLIT_1_3)

    def test_to_df(self, LIVESPLIT_1_7_0):
        splits = ss.read_lss(LIVESPLIT_1_7_0)
        splits_df = splits.to_df(
            time_type=TimeType.REAL_TIME,
            allow_partial=True,
            allow_empty=True,
            cumulative=False,
            lss_repr=False,
        )
        assert set(splits_df.columns) == set(splits._collect_ids())

    def test_to_df_cumulative(self, LIVESPLIT_1_7_0):
        splits = ss.read_lss(LIVESPLIT_1_7_0)
        cumulative_splits = splits.to_df(
            time_type=TimeType.REAL_TIME, allow_partial=False, cumulative=True, lss_repr=False
        )

        isolated_splits = splits.to_df(
            time_type=TimeType.REAL_TIME, allow_partial=False, cumulative=False, lss_repr=False
        )

        assert np.array_equal(cumulative_splits.iloc[-1, :].values, isolated_splits.sum().values)



from packaging.version import Version

DATETIME_FORMAT = r"%m/%d/%Y %H:%M:%S"
NANOSECONDS_DAY = 86400 * 10**9
NANOSECONDS_HOUR = 3600 * 10**9
NANOSECONDS_MINUTE = 60 * 10**9
NANOSECONDS_SECOND = 10**9
MINIMUM_LSS_VERSION = Version("1.6.0")



from __future__ import annotations
from pandas import Timedelta
from pydantic import conint
from pydantic_xml import BaseXmlModel, attr, element, wrapped
from typing import List, Optional
from saltysplits.annotations import TimeOptional, DateTime, SBool, OffsetOptional, LssVersion


class Splits(BaseXmlModel, tag="Run", arbitrary_types_allowed=True, search_mode="ordered"):
    version: LssVersion = attr(name="version")
    game_icon: Optional[str] = element(tag="GameIcon", default=None)
    game_name: str = element(tag="GameName")
    category_name: str = element(tag="CategoryName")
    layout_path: Optional[str] = element(tag="LayoutPath", default=None)
    offset: OffsetOptional = element(tag="Offset", default=Timedelta(0))
    attempt_count: Optional[conint(ge=0)] = element(tag="AttemptCount", default=0)
    attempt_history: Optional[List[Attempt]] = wrapped("AttemptHistory", default=None)
    segments: List[Segment] = wrapped("Segments")


class BaseTime(BaseXmlModel, arbitrary_types_allowed=True, search_mode="ordered"):
    real_time: TimeOptional = element(tag="RealTime", default=None)
    game_time: TimeOptional = element(tag="GameTime", default=None)


class Attempt(BaseTime, tag="Attempt"):
    id: str = attr(name="id")
    started: Optional[DateTime] = attr(name="started", default=None)
    is_started_synced: Optional[SBool] = attr(name="isStartedSynced", default=None)
    ended: Optional[DateTime] = attr(name="ended", default=None)
    is_ended_synced: Optional[SBool] = attr(name="isEndedSynced", default=None)


class Segment(BaseXmlModel, tag="Segment", arbitrary_types_allowed=True, search_mode="ordered"):
    name: str = element(tag="Name")
    icon: Optional[str] = element(tag="Icon", default=None)
    split_times: List[SplitTime] = wrapped("SplitTimes")
    best_segment_time: BaseTime = element(tag="BestSegmentTime")
    segment_history: Optional[List[Time]] = wrapped("SegmentHistory", default=None)


class SplitTime(BaseTime, tag="SplitTime"):
    name: str = attr(name="name")


class Time(BaseTime, tag="Time"):
    id: str = attr(name="id")


# ensures all model elements are defined when we need them (without relying on arbritrary definition order)
Splits.model_rebuild()
Segment.model_rebuild()



import pathlib
from .main import SaltySplits
from .enums import TimeType as TimeType
from functools import partial as _partial

DEMO_SPLITS = pathlib.Path(__file__).parents[2] / "tests/run_files/gcb.lss"
LOGO_PATH = pathlib.Path(__file__).parents[2] / "docs/assets/images/logo.png"

# forwarding read_lss to main module (+ docs)
read_lss = _partial(SaltySplits.read_lss)
read_lss.__doc__ = SaltySplits.read_lss.__doc__



import pytest  # noqa: F401
import lxml.etree as ET
from saltysplits.models import Splits, Attempt, Segment, SplitTime, Time
from .conftest import drop_empty_tags


class TestSplits:
    def test_instancing(self, LIVESPLIT_1_7_0):
        tree = ET.parse(LIVESPLIT_1_7_0)
        element = tree.getroot()

        # we currently don't model MetaData and AutosplitterSettings so we drop them before comparison
        element.remove(element.find("Metadata"))
        element.remove(element.find("AutoSplitterSettings"))

        # dropping empty tags before comparison, no way to catch that AND actually optional elements (e.g. real_time and/or game_time)
        _ = drop_empty_tags(element=element, top_level=False)
        element_string = ET.tostring(element, method="c14n2", strip_text=True).decode("utf-8")

        # validating if we can instance from known valid XML element
        model = Splits.from_xml_tree(element)
        recreated_element = model.to_xml_tree(
            exclude_unset=True
        )  # only include elements actually set by the XML
        recreated_element_string = ET.tostring(
            recreated_element, method="c14n2", strip_text=True
        ).decode("utf-8")

        # validating that encoded model is identical to original XML element
        assert element_string == recreated_element_string


class TestAttempt:
    def test_dump(self, LIVESPLIT_1_7_0):
        tree = ET.parse(LIVESPLIT_1_7_0)
        root = tree.getroot()
        element = next(root.iter("Attempt"))
        # dropping empty tags before comparison, no way to catch that AND actually optional elements (e.g. real_time and/or game_time)
        _ = drop_empty_tags(element=element)
        element_string = ET.tostring(element, method="c14n2", strip_text=True)

        # validating if we can instance from known valid XML element
        model = Attempt.from_xml_tree(element)
        recreated_element = model.to_xml_tree(
            exclude_unset=True
        )  # only include elements actually set by the XML
        recreated_element_string = ET.tostring(recreated_element, method="c14n2", strip_text=True)

        # validating that encoded model is identical to original XML element
        assert element_string == recreated_element_string


class TestSegment:
    def test_dump(self, LIVESPLIT_1_7_0):
        tree = ET.parse(LIVESPLIT_1_7_0)
        root = tree.getroot()
        element = next(root.iter("Segment"))
        # dropping empty tags before comparison, no way to catch that AND actually optional elements (e.g. real_time and/or game_time)
        _ = drop_empty_tags(element=element)
        element_string = ET.tostring(element, method="c14n2", strip_text=True)

        # validating if we can instance from known valid XML element
        model = Segment.from_xml_tree(element)
        recreated_element = model.to_xml_tree(
            exclude_unset=True
        )  # only include elements actually set by the XML
        recreated_element_string = ET.tostring(recreated_element, method="c14n2", strip_text=True)

        # validating that encoded model is identical to original XML element
        assert element_string == recreated_element_string


class TestSplitTime:
    def test_dump(self, LIVESPLIT_1_7_0):
        tree = ET.parse(LIVESPLIT_1_7_0)
        root = tree.getroot()
        element = next(root.iter("SplitTime"))
        # dropping empty tags before comparison, no way to catch that AND actually optional elements (e.g. real_time and/or game_time)
        _ = drop_empty_tags(element=element)
        element_string = ET.tostring(element, method="c14n2", strip_text=True)

        # validating if we can instance from known valid XML element
        model = SplitTime.from_xml_tree(element)
        recreated_element = model.to_xml_tree(
            exclude_unset=True
        )  # only include elements actually set by the XML
        recreated_element_string = ET.tostring(recreated_element, method="c14n2", strip_text=True)

        # validating that encoded model is identical to original XML element
        assert element_string == recreated_element_string


class TestTime:
    def test_dump(self, LIVESPLIT_1_7_0):
        tree = ET.parse(LIVESPLIT_1_7_0)
        root = tree.getroot()
        element = next(root.iter("Time"))
        # dropping empty tags before comparison, no way to catch that AND actually optional elements (e.g. real_time and/or game_time)
        _ = drop_empty_tags(element=element)
        element_string = ET.tostring(element, method="c14n2", strip_text=True)

        # validating if we can instance from known valid XML element
        model = Time.from_xml_tree(element)
        recreated_element = model.to_xml_tree(
            exclude_unset=True
        )  # only include elements actually set by the XML
        recreated_element_string = ET.tostring(recreated_element, method="c14n2", strip_text=True)

        # validating that encoded model is identical to original XML element
        assert element_string == recreated_element_string
